unit UdNotes;

interface

uses
  System.Classes;

procedure Register;

implementation

uses
  Winapi.Windows, System.SysUtils, System.Generics.Collections, System.Rtti, System.TypInfo,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.Menus, ToolsAPI;

const
  {$IFDEF VER210}vclide = 'vclide140.bpl';{$ENDIF} // 2010
  {$IFDEF VER220}vclide = 'vclide150.bpl';{$ENDIF} // XE
  {$IFDEF VER230}vclide = 'vclide160.bpl';{$ENDIF} // XE2
  {$IFDEF VER240}vclide = 'vclide170.bpl';{$ENDIF} // XE3
  {$IFDEF VER250}vclide = 'vclide180.bpl';{$ENDIF} // XE4
  {$IFDEF VER260}vclide = 'vclide190.bpl';{$ENDIF} // XE5
  {$IFDEF VER270}vclide = 'vclide200.bpl';{$ENDIF} // XE6
  {$IFDEF VER280}vclide = 'vclide210.bpl';{$ENDIF} // XE7
  {$IFDEF VER290}vclide = 'vclide220.bpl';{$ENDIF} // XE8

type
  PVirtualNode = Pointer; // = ^TVirtualNode
  TColumnIndex = type Integer;

  TVTColumnOption = (coAllowClick, coDraggable, coEnabled, coParentBidiMode,
    coParentColor, coResizable, coShowDropMark, coVisible, coAutoSpring, coFixed);
  TVTColumnOptions = set of TVTColumnOption;

  TVTHeaderOption = (hoAutoResize, hoColumnResize, hoDblClickResize, hoDrag,
    hoHotTrack, hoOwnerDraw, hoRestrictDrag, hoShowHint, hoShowImages,
  hoShowSortGlyphs, hoVisible, hoAutoSpring);
  TVTHeaderOptions = set of TVTHeaderOption;

  TVirtualTreeColumn = class(TCollectionItem)
  end;

  TVirtualTreeColumns = class(TCollection)
  private
    function GetItem(Index: TColumnIndex): TVirtualTreeColumn;
  public
    function Add: TVirtualTreeColumn;
    property Items[Index: TColumnIndex]: TVirtualTreeColumn read GetItem; default;
  end;

  TBaseVirtualTree = class
  public
    function GetNodeData(Node: PVirtualNode): Pointer;
    function GetNodeLevel(Node: PVirtualNode): Cardinal;
    function GetNodeParent(Node: PVirtualNode): PVirtualNode;
  end;

  TUnitComments = class
  private
  var
    FFileName: string;
    FItems: TDictionary<string,string>;
    function FileNameToKey(const AFileName: string): string;
    function GetComment(const AUnitFileName: string): string;
    procedure SetComment(const AUnitFileName, Value: string);
  public
    constructor Create(const AFileName: string);
    destructor Destroy; override;
    procedure Load;
    procedure Save;
    procedure RenameUnit(const AOldFileName, ANewFileName: string);
    procedure RemoveUnit(const AUnitFileName: string);
    property Comment[const AUnitFileName: string]: string read GetComment write SetComment; default;
  end;

  TProjects = class
  private
  const
    UnitCommentExt = '.dnotes';
  var
    FProjects: TObjectDictionary<string,TUnitComments>;
    function FileNameToKey(const AFileName: string): string;
    function GetComment(const ABaseFileName, AUnitFileName: string): string;
    procedure SetComment(const ABaseFileName, AUnitFileName, Value: string);
  public
    constructor Create;
    destructor Destroy; override;
    procedure LoadComments(const ABaseFileName: string);
    procedure SaveComments(const ABaseFileName: string);
    procedure CloseComments(const ABaseFileName: string);
    procedure RenameProject(const OldFileName, NewFileName: string);
    procedure RenameUnit(const ABaseFileName, OldFileName, NewFileName: string);
    procedure RemoveUnit(const ABaseFileName, AUnitFileName: string);
    property Comment[const ABaseFileName, AUnitFileName: string]: string read GetComment write SetComment; default;
  end;

  TProjectMenuCreatorNotifier = class(TNotifierObject, INTAProjectMenuCreatorNotifier)
  private
    function AddMenu(const Ident: string): TMenuItem;
    function CanHandle(const Ident: string): Boolean;
    procedure MenuClick(Sender: TObject);
  end;

  TIDENotifier = class;

  IOTAProjectNotifierEx = interface(IOTAProjectNotifier)
    function GetModule: IOTAModule;
    property Module: IOTAModule read GetModule;
  end;

  TProjectNotifier = class(TInterfacedObject, IOTAModuleNotifier, IOTAProjectNotifier, IOTAProjectNotifierEx)
  private
    FOwner: TIDENotifier;
    FModule: IOTAModule;
    FNotifierIndex: Integer;
    FFileName: string;

    // IOTANotifier
    procedure AfterSave;
    procedure BeforeSave;
    procedure Destroyed;
    procedure Modified;
    // IOTAModuleNotifier
    function CheckOverwrite: Boolean;
    procedure ModuleRenamed(const NewName: string); overload;
    // IOTAProject
    procedure ModuleAdded(const AFileName: string);
    procedure ModuleRemoved(const AFileName: string);
    procedure ModuleRenamed(const AOldFileName, ANewFileName: string); overload;
    // IOTAProjectNotifierEx
    function GetModule: IOTAModule;
  public
    constructor Create(AOwner: TIDENotifier; const AModule: IOTAModule);
    destructor Destroy; override;
  end;

  TIDENotifier = class(TNotifierObject, IOTAIDENotifier)
  private
    FModuleNotifiers: TList<IOTAProjectNotifierEx>;
    procedure FileNotification(NotifyCode: TOTAFileNotification;
      const FileName: string; var Cancel: Boolean);
    procedure BeforeCompile(const Project: IOTAProject; var Cancel: Boolean);
    procedure AfterCompile(Succeeded: Boolean);
    function IsModuleNotifierInstalled(const AModule: IOTAModule): Boolean;
    procedure InstallModuleNotifiers;
  public
    constructor Create;
    destructor Destroy; override;
  end;

  TVSTGetTextEvent_Wide = procedure(Sender: TBaseVirtualTree; Node: PVirtualNode;
    Column: TColumnIndex; TextType: Byte{TVSTTextType};
    var CellText: WideString) of object;

const
  TVSTGetTextEvent_ParameterNames_Wide: array[0..4] of string = (
    'TBaseVirtualTree', 'PVirtualNode', 'TColumnIndex', 'TVSTTextType', 'WideString'
  );

var
  Org_OnGetText_Wide: TVSTGetTextEvent_Wide;
  Projects: TProjects;
  PropTStdFileContainer_FileName: PPropInfo;
  ClsTStdFileContainer: TClass;

{ Inner classes }

function TVirtualTreeColumns.Add: TVirtualTreeColumn; external vclide name '@Idevirtualtrees@TVirtualTreeColumns@Add$qqrv';
function TVirtualTreeColumns.GetItem(Index: TColumnIndex): TVirtualTreeColumn; external vclide name '@Idevirtualtrees@TVirtualTreeColumns@GetItem$qqri';
function TBaseVirtualTree.GetNodeData(Node: Pointer): Pointer; external vclide name '@Idevirtualtrees@TBaseVirtualTree@GetNodeData$qqrp28Idevirtualtrees@TVirtualNode';
function TBaseVirtualTree.GetNodeLevel(Node: PVirtualNode): Cardinal; external vclide name '@Idevirtualtrees@TBaseVirtualTree@GetNodeLevel$qqrp28Idevirtualtrees@TVirtualNode';
function TBaseVirtualTree.GetNodeParent(Node: PVirtualNode): PVirtualNode; external vclide name '@Idevirtualtrees@TBaseVirtualTree@GetNodeParent$qqrp28Idevirtualtrees@TVirtualNode';

{ TUnitComments }

constructor TUnitComments.Create(const AFileName: string);
begin
  inherited Create;
  FFileName := AFileName;
  FItems := TDictionary<string,string>.Create;
end;

destructor TUnitComments.Destroy;
begin
  FItems.Free;
  inherited;
end;

function TUnitComments.FileNameToKey(const AFileName: string): string;
begin
  Result := ExtractFileName(AFileName);
end;

procedure TUnitComments.Load;
var
  key: string;
  sl: TStringList;
  i: Integer;
begin
  if not FileExists(FFileName) then Exit;

  sl := TStringList.Create;
  try
    sl.LoadFromFile(FFileName, TEncoding.UTF8);
    for i := 0 to sl.Count-1 do
    begin
      key := sl.Names[i];
      if not FItems.ContainsKey(key) then
        FItems.Add(key, sl.ValueFromIndex[i]);
    end;
  finally
    sl.Free;
  end;
end;

procedure TUnitComments.Save;
var
  key: string;
  sl: TStringList;
begin
  sl := TStringList.Create;
  try
    for key in FItems.Keys do
      if FItems[key] <> '' then
        sl.Add(key + '=' + FItems[key]);

    if sl.Count > 0 then
      sl.SaveToFile(FFileName, TEncoding.UTF8)
    else begin
      if FileExists(FFileName) then
        DeleteFile(FFileName);
    end;
  finally
    sl.Free;
  end;
end;

procedure TUnitComments.RenameUnit(const AOldFileName, ANewFileName: string);
var
  key, value: string;
begin
  key := FileNameToKey(AOldFileName);
  if not FItems.ContainsKey(key) then Exit;
  value := FItems.ExtractPair(key).Value;
  FItems.Add(FileNameToKey(ANewFileName), value);
end;

procedure TUnitComments.RemoveUnit(const AUnitFileName: string);
begin
  FItems.Remove(FileNameToKey(AUnitFileName));
end;

function TUnitComments.GetComment(const AUnitFileName: string): string;
begin
  if not FItems.TryGetValue(FileNameToKey(AUnitFileName), Result) then
    Result := '';
end;

procedure TUnitComments.SetComment(const AUnitFileName, Value: string);
begin
  FItems.AddOrSetValue(FileNameToKey(AUnitFileName), Value);
end;

{ TProjects }

constructor TProjects.Create;
begin
  inherited;
  FProjects := TObjectDictionary<string,TUnitComments>.Create([doOwnsValues]);
end;

destructor TProjects.Destroy;
begin
  FProjects.Free;
  inherited;
end;

function TProjects.FileNameToKey(const AFileName: string): string;
begin
  Result := ChangeFileExt(AFileName, UnitCommentExt);
end;

procedure TProjects.LoadComments(const ABaseFileName: string);
var
  prj: TUnitComments;
  key: string;
begin
  key := FileNameToKey(ABaseFileName);
  if FProjects.TryGetValue(key, prj) then
    prj.FItems.Clear
  else begin
    prj := TUnitComments.Create(key);
    FProjects.Add(key, prj);
  end;
  prj.Load;
end;

procedure TProjects.SaveComments(const ABaseFileName: string);
var
  key: string;
  prj: TUnitComments;
begin
  key := FileNameToKey(ABaseFileName);
  if not FProjects.TryGetValue(key, prj) then Exit;
  prj.Save;
end;

procedure TProjects.CloseComments(const ABaseFileName: string);
var
  key: string;
begin
  key := FileNameToKey(ABaseFileName);
  SaveComments(ABaseFileName);
  FProjects.Remove(key);
end;

procedure TProjects.RenameProject(const OldFileName, NewFileName: string);
var
  prj: TUnitComments;
begin
  prj := FProjects.ExtractPair(FileNameToKey(OldFileName)).Value;
  if prj = nil then Exit;

  prj.FFileName := FileNameToKey(NewFileName);
  FProjects.Add(FileNameToKey(NewFileName), prj);

  if FileExists(FileNameToKey(OldFileName)) then
    MoveFileEx(PChar(FileNameToKey(OldFileName)), PChar(FileNameToKey(NewFileName)), MOVEFILE_REPLACE_EXISTING or MOVEFILE_WRITE_THROUGH);
end;

procedure TProjects.RenameUnit(const ABaseFileName, OldFileName,
  NewFileName: string);
var
  prj: TUnitComments;
begin
  if not FProjects.TryGetValue(FileNameToKey(ABaseFileName), prj) then Exit;

  prj.RenameUnit(OldFileName, NewFileName);
end;

procedure TProjects.RemoveUnit(const ABaseFileName, AUnitFileName: string);
var
  prj: TUnitComments;
begin
  if not FProjects.TryGetValue(FileNameToKey(ABaseFileName), prj) then Exit;

  prj.RemoveUnit(AUnitFileName);
end;

function TProjects.GetComment(const ABaseFileName, AUnitFileName: string): string;
var
  prj: TUnitComments;
begin
  if not FProjects.TryGetValue(FileNameToKey(ABaseFileName), prj) then Exit('');

  Result := prj[AUnitFileName];
end;

procedure TProjects.SetComment(const ABaseFileName, AUnitFileName, Value: string);
var
  prj: TUnitComments;
  key: string;
begin
  key := FileNameToKey(ABaseFileName);
  if not FProjects.TryGetValue(key, prj) then
  begin
    prj := TUnitComments.Create(key);
    FProjects.Add(key, prj);
  end;
  prj[AUnitFileName] := Value;
end;

{ TProjectMenuCreatorNotifier }

function TProjectMenuCreatorNotifier.AddMenu(const Ident: string): TMenuItem;
begin
  Result := TMenuItem.Create(nil);
  Result.Caption := 'Rg';
  Result.OnClick := MenuClick;
end;

function TProjectMenuCreatorNotifier.CanHandle(const Ident: string): Boolean;
var
  ext: string;
begin
  ext := ExtractFileExt(Ident);
  Result := (ext = '.pas') or (ext = '.dfm') or (ext = '.fmx');
end;

function GetProjectTree: TControl; forward;

procedure TProjectMenuCreatorNotifier.MenuClick(Sender: TObject);
var
  project: IOTAProject;
  ident: string;
  ext: string;
  comment: string;
begin
  project := (BorlandIDEServices as IOTAProjectManager).GetCurrentSelection(ident);
  ext := ExtractFileExt(ident).ToLower;
  if (ext <> '.pas') and (ext <> '.dfm') and (ext <> '.fmx') then Exit;

  comment := Projects[project.FileName, ident];

  if not InputQuery('Rg', 'Rg:', comment) then Exit;

  Projects[project.FileName, ident] := comment;
  GetProjectTree.Invalidate;
end;

{ TProjectNotifier }

constructor TProjectNotifier.Create(AOwner: TIDENotifier; const AModule: IOTAModule);
begin
  inherited Create;
  FOwner := AOwner;
  FNotifierIndex := -1;
  FModule := AModule;
  FNotifierIndex := FModule.AddNotifier(Self);
  FOwner.FModuleNotifiers.Add(Self);
  FFileName := FModule.FileName;
  Projects.LoadComments(FFileName);
end;

destructor TProjectNotifier.Destroy;
begin
  Projects.CloseComments(FFileName);
  if FNotifierIndex >= 0 then
  begin
    FModule.RemoveNotifier(FNotifierIndex);
    FOwner.FModuleNotifiers.Remove(Self);
  end;
  inherited;
end;

procedure TProjectNotifier.AfterSave;
begin
  { Do nothing }
end;

procedure TProjectNotifier.BeforeSave;
begin
  { Do nothing }
end;

procedure TProjectNotifier.Destroyed;
begin
  if FNotifierIndex >= 0 then
  begin
    FModule.RemoveNotifier(FNotifierIndex);
    FNotifierIndex := -1;
    FOwner.FModuleNotifiers.Remove(Self);
  end;
  FModule := nil;
end;

procedure TProjectNotifier.Modified;
begin
  { Do nothing }
end;

function TProjectNotifier.CheckOverwrite: Boolean;
begin
  Result := True;
end;

procedure TProjectNotifier.ModuleRenamed(const AOldFileName,
  ANewFileName: string);
var
  i: Integer;
  module: IOTAModule;
begin
  if AOldFileName = ANewFileName then Exit;

  module := (BorlandIDEServices as IOTAModuleServices).FindModule(AOldFileName);
  if module = nil then Exit;

  for i := 0 to module.OwnerCount-1 do
    Projects.RenameUnit(module.Owners[i].FileName, AOldFileName, ANewFileName);
end;

procedure TProjectNotifier.ModuleAdded(const AFileName: string);
begin
  { Do nothing }
end;

procedure TProjectNotifier.ModuleRemoved(const AFileName: string);
begin
  Projects.RemoveUnit(FFileName, AFileName);
end;

procedure TProjectNotifier.ModuleRenamed(const NewName: string);
begin
  if FFileName = NewName then Exit;

  Projects.RenameProject(FFileName, NewName);
  FFileName := NewName;
end;

function TProjectNotifier.GetModule: IOTAModule;
begin
  Result := FModule;
end;

{ TIDENotifier }

constructor TIDENotifier.Create;
begin
  inherited;
  FModuleNotifiers := TList<IOTAProjectNotifierEx>.Create;
  InstallModuleNotifiers;
end;

destructor TIDENotifier.Destroy;
begin
  while FModuleNotifiers.Count > 0 do
    FModuleNotifiers[FModuleNotifiers.Count-1].Destroyed;
  FModuleNotifiers.Free;
  inherited;
end;

procedure TIDENotifier.FileNotification(NotifyCode: TOTAFileNotification;
  const FileName: string; var Cancel: Boolean);
begin
  if NotifyCode = ofnFileOpened then
    InstallModuleNotifiers;
end;

procedure TIDENotifier.BeforeCompile(const Project: IOTAProject;
  var Cancel: Boolean);
begin
  { Do nothing }
end;

procedure TIDENotifier.AfterCompile(Succeeded: Boolean);
begin
  { Do nothing }
end;

function TIDENotifier.IsModuleNotifierInstalled(const AModule: IOTAModule): Boolean;
var
  i: Integer;
begin
  for i := 0 to FModuleNotifiers.Count-1 do
    if FModuleNotifiers[i].Module = AModule then Exit(True);
  Result := False;
end;

procedure TIDENotifier.InstallModuleNotifiers;
var
  moduleServices: IOTAModuleServices;
  module: IOTAModule;
  i: Integer;
begin
  moduleServices := BorlandIDEServices as IOTAModuleServices;
  for i := 0 to moduleServices.ModuleCount-1 do
  begin
    module := moduleServices.Modules[i];
    if Supports(module, IOTAProject) and not Supports(module, IOTAProjectGroup) then
    begin
      if not IsModuleNotifierInstalled(module) then
        TProjectNotifier.Create(Self, module);
    end;
  end;
end;

{ Functions }

procedure VSTGetText_Wide(Self: TObject; Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex;
  TextType: Byte; var CellText: WideString);
var
  intf: ^IInterface;
  fname, ext: string;
  obj: TObject;

  function GetNodeFileName(Node: PVirtualNode): string;
  var
    intf: ^IInterface;
    obj: TObject;
  begin
    intf := Sender.GetNodeData(Node);
    obj := intf^ as TObject;
    if obj is ClsTStdFileContainer then
      Result := GetStrProp(obj, PropTStdFileContainer_FileName)
    else
      Result := '';
  end;

  function GetProjectFileName(Node: PVirtualNode): string;
  begin
    while Sender.GetNodeLevel(Node) > 1 do
      Node := Sender.GetNodeParent(Node);
    Result := GetNodeFileName(Node);
  end;

begin
  if Column <> 2 then
  begin
    if Assigned(Org_OnGetText_Wide) then
      Org_OnGetText_Wide(Sender, Node, Column, TextType, CellText);
    Exit;
  end;

  if ClsTStdFileContainer = nil then Exit;
  if PropTStdFileContainer_FileName = nil then Exit;
  try
    intf := Sender.GetNodeData(Node);
    obj := intf^ as TObject;
    if obj is ClsTStdFileContainer then
    begin
      fname := GetNodeFileName(Node);
      ext := ExtractFileExt(fname).ToLower;

      if (ext <> '.pas') and (ext <> '.dfm') and (ext <> '.fmx') then Exit;

      CellText := Projects[GetProjectFileName(Node), fname];
    end;
  except
  end;
end;

function GetProjectTree: TControl;
var
  i, j: Integer;
begin
  for i := 0 to Screen.FormCount-1 do
    if SameText(Screen.Forms[i].ClassName, 'TProjectManagerForm') then
    begin
      Result := Screen.Forms[i].FindChildControl('ProjectTree2');
      if Result <> nil then Exit;

      for j := 0 to Screen.Forms[i].ControlCount-1 do
        if SameText(Screen.Forms[i].Controls[j].ClassName, 'TVirtualStringTree') then
          Exit(Screen.Forms[i].Controls[j]);

      Break;
    end;
  Result := nil;
end;

function SetProjectTreeOnGetText(Tree: TObject; OnGetText: TMethod): TMethod;
var
  ctx: TRttiContext;
  typ: TRttiType;
  prop: TRttiProperty;
  params: TArray<TRttiParameter>;
  i: Integer;
begin
  typ := ctx.GetType(Tree.ClassType);
  prop := typ.GetProperty('OnGetText');
  if TRttiMethodType(prop.PropertyType).MethodKind <> mkProcedure then Exit;

  params := TRttiMethodType(prop.PropertyType).GetParameters;
  if Length(params) <> Length(TVSTGetTextEvent_ParameterNames_Wide) then Exit;
  for i := Low(params) to High(params) do
    if not SameText(params[i].ParamType.Name, TVSTGetTextEvent_ParameterNames_Wide[i]) then
      Exit;

  Result := GetMethodProp(Tree, TRttiInstanceProperty(prop).PropInfo);
  SetMethodProp(Tree, TRttiInstanceProperty(prop).PropInfo, OnGetText);
end;

procedure VTHeader_Options_SetAutoResize(Header: TObject; Value: Boolean);
var
  i: Integer;
begin
  i := GetOrdProp(Header, 'Options');
  if Value then
    Include(TIntegerSet(i), GetEnumValue(TypeInfo(TVTHeaderOption), 'hoAutoResize'))
  else
    Exclude(TIntegerSet(i), GetEnumValue(TypeInfo(TVTHeaderOption), 'hoAutoResize'));
  SetOrdProp(Header, 'Options', i);
end;

procedure VTColumn_Options_SetAutoSpring(Column: TObject; Value: Boolean);
var
  i: Integer;
begin
  i := GetOrdProp(Column, 'Options');
  if Value then
    Include(TIntegerSet(i), GetEnumValue(TypeInfo(TVTColumnOption), 'coAutoSpring'))
  else
    Exclude(TIntegerSet(i), GetEnumValue(TypeInfo(TVTColumnOption), 'coAutoSpring'));
  SetOrdProp(Column, 'Options', i);
end;

procedure InitRTTICache;
var
  ctx: TRttiContext;
  typ: TRttiType;
  prop: TRttiProperty;
begin
  typ := ctx.FindType('Containers.TStdFileContainer');
  if typ <> nil then
  begin
    ClsTStdFileContainer := TRttiInstanceType(typ).MetaclassType;

    prop := typ.GetProperty('FileName');
    if prop <> nil then
      PropTStdFileContainer_FileName := TRttiInstanceProperty(prop).PropInfo;
  end;
end;

var
  IDENotifierIndex: Integer = -1;
  MenuNotifierIndex: Integer = -1;

procedure Register;
var
  objTree, objHeader: TObject;
  objColumns: TVirtualTreeColumns;
  objColumn: TVirtualTreeColumn;
  method: TMethod;
begin
  InitRTTICache;
  Projects := TProjects.Create;

  objTree := GetProjectTree;
  if objTree = nil then Exit;

  objHeader := GetObjectProp(objTree, 'Header');
  if objHeader = nil then Exit;

  objColumns := TVirtualTreeColumns(GetObjectProp(objHeader, 'Columns'));
  if objColumns = nil then Exit;

  objColumn := objColumns.Add;
  VTColumn_Options_SetAutoSpring(objColumn, True);

  SetStrProp(objColumn, 'Text', 'Rg');

  method.Code := @VSTGetText_Wide;
  method.Data := nil;
  Org_OnGetText_Wide := TVSTGetTextEvent_Wide(SetProjectTreeOnGetText(objTree, method));

  VTHeader_Options_SetAutoResize(objHeader, False);

  IDENotifierIndex := (BorlandIDEServices as IOTAServices).AddNotifier(TIDENotifier.Create);
  // TODO: ȂꍇAddMenuItemCreatorNotifiergɕς
  MenuNotifierIndex := (BorlandIDEServices as IOTAProjectManager).AddMenuCreatorNotifier(TProjectMenuCreatorNotifier.Create);
end;

procedure Unregister;
var
  objTree, objHeader: TObject;
  objColumns: TVirtualTreeColumns;
  i: Integer;
begin
  objTree := GetProjectTree;
  if objTree <> nil then
  begin
    objHeader := GetObjectProp(objTree, 'Header');
    if objHeader <> nil then
    begin
      objColumns := TVirtualTreeColumns(GetObjectProp(objHeader, 'Columns'));
      if objColumns <> nil then
      begin
        for i := objColumns.Count-1 downto 2 do
          objColumns.Delete(i);

        SetProjectTreeOnGetText(objTree, TMethod(Org_OnGetText_Wide));

        // TODO: pX(Columns[1])\ĂꍇFalsê܂܂ɂ
        VTHeader_Options_SetAutoResize(objHeader, True);
      end;
    end;
  end;

  if IDENotifierIndex >= 0 then
  try
    (BorlandIDEServices as IOTAServices).RemoveNotifier(IDENotifierIndex);
  except
  end;
  if MenuNotifierIndex >= 0 then
    (BorlandIDEServices as IOTAProjectManager).RemoveMenuCreatorNotifier(MenuNotifierIndex);
  Projects.Free;
end;

initialization
finalization
  Unregister;
end.
